
import sys
import clingo

import selftest
test = selftest.get_tester(__name__)


def clingo_control_plugin(next, control=None, arguments=[], message_limit=20, **etc):
    """ Provides a default control when there is none. """
    
    def logger(code, message):
        _logger(code, message)
    
    if not control:
        control = clingo.Control(logger=logger, arguments=arguments, message_limit=message_limit)
        
    _logger, main = next(logger=logger, control=control, arguments=arguments, message_limit=message_limit, **etc)
        
    return main


@test
def control_plugin_basics():
    trace = []
    def next(control=None, more=None, **etc):
        test.eq('better', more)
        trace.append(control)
        def logger(code, message):
            trace.append(message)
        def main():
            return 43
        return logger, main
    main = clingo_control_plugin(next, more='better')
    control = trace[0]
    test.isinstance(control, clingo.Control)
    test.eq(1, len(trace))
    try:
        control.add("error")
    except RuntimeError:
        pass
    test.eq('<block>:2:1-2: error: syntax error, unexpected EOF\n', trace[1])
    r = main()
    test.eq(43, r)


@test
def control_plugin_makes_no_control_when_given():
    trace = []
    def next(control=None, **etc):
        trace.append(control)
        return 1, 2
    main = clingo_control_plugin(next, control="ceçi c'est un Control")
    test.eq("ceçi c'est un Control", trace[0])
    test.eq(2, main)
    test.eq(1, len(trace))


@test
def pass_arguments_to_control():
    trace = []
    def next_plugin(control=None, arguments=(), **etc):
        trace.append(control)
        trace.append(arguments)
        def logger(code, message):
            trace.append(message)
        return logger, lambda: None
    clingo_control_plugin(next_plugin, arguments=['--const', 'a=42'])
    test.eq([test.any, ['--const', 'a=42']], trace)
    control = trace[0]
    control.add("b(a * 2).")
    control.ground()
    test.eq([test.any, ['--const', 'a=42']], trace)
    b = next(control.symbolic_atoms.by_signature('b', 1))
    test.eq('b(84)', str(b.symbol))


@test
def pass_arguments_to_control():
    trace = []
    logs = []
    def next_plugin(control=None, message_limit=20, **etc):
        trace.append(control)
        trace.append(message_limit)
        def logger(code, message):
            logs.append(message)
        return logger, lambda: None
    clingo_control_plugin(next_plugin, message_limit=2)
    test.eq([test.any, 2], trace)
    control = trace[0]
    control.add("p :- a.  q :- b.  r :- c.")  # cause 3 messages
    control.ground()
    test.eq([                                 # only 2 get logged
        '<block>:1:6-7: info: atom does not occur in any rule head:\n  a\n',
        '<block>:1:15-16: info: atom does not occur in any rule head:\n  b\n'],
        logs)




import tempfile
import pathlib

import selftest
test = selftest.get_tester(__name__)

from .misc import write_tempfile


def source_plugin(next, source=None, label=None, files=(), **etc):
    """ Turns source as string into temporary file. """
    
    source_file = None
            
    if source:
        source_file = write_tempfile(f"-{label or 'string'}.lp", source)
        files = (*files, source_file.name)

    _main = next(files=files, **etc)
    
    def main():
        try:
            return _main()
        finally:
            if source_file:
                source_file.close()
            
    return main


@test
def source_plugin_basics():
    
    trace = []
        
    def next_plugin(files=(), misc=None):
        trace.append(files)
        trace.append(misc)
        def next_main():
            trace.append('next_main')
        return next_main
            
    main = source_plugin(
        next_plugin,
        source="one(1).",
        label='my-logic',
        files=('a.lp',),
        misc='fortytwo')

    test.eq(('a.lp', test.any), trace[0])
    test.eq('fortytwo', trace[1])
    filename = trace[0][1]
    test.endswith(filename, "-my-logic.lp")
    test.truth(pathlib.Path(filename).exists())
    test.eq("one(1).", open(filename).read())
    test.eq(2, len(trace))

    main()
    test.comp.truth(pathlib.Path(filename).exists())
    test.eq('next_main', trace[2])
    test.eq(3, len(trace))


@test
def no_source_given():

    trace = []
        
    def next_plugin(files=(), extra=None):
        trace.append(files)
        trace.append(extra)
        def next_main():
            trace.append('next_main')
            return 42
        return next_main
            
    main = source_plugin(
        next_plugin,
        files=('b.lp',),
        extra='more')

    test.eq(('b.lp',), trace[0])
    test.eq('more', trace[1])
    test.eq(2, len(trace))
    r = main()
    test.eq(42, r)
    test.eq('next_main', trace[2])
    test.eq(3, len(trace))


@test
def avoid_None_as_label():
    trace = []
    def next_plugin(source='no-source', label='no-label', files=()):
        trace.append(source)
        trace.append(label)
        trace.append(files)
    main = source_plugin(next_plugin, source="a.")
    test.eq('no-source', trace[0])
    test.eq('no-label', trace[1])
    test.endswith(trace[2][0], '-string.lp')





""" Integration tests """


import sys
import os

import clingo
from clingo.script import enable_python
enable_python()

from .plugins.misc import write_file, list_symbols

from .session2 import session2, clingo_session

from .plugins import (
    source_plugin,
    clingo_control_plugin,
    clingo_sequencer_plugin,
    insert_plugin_plugin,
    clingo_defaults_plugin,
    clingo_syntaxerror_plugin,
    testrunner_plugin,
    stdin_to_tempfile_plugin,
    compound_context_plugin,
)

import selftest
test = selftest.get_tester(__name__)


def ground_exc(source=None, label=None, files=(), include_paths=(), arguments=(), observer=None, trace=None, **etc):
    """ a general pupose one-stop ground function """
    old_path = os.environ.get('CLINGOPATH', '').split(':')
    os.environ['CLINGOPATH'] = ':'.join((*old_path, *include_paths))
    try:
        control = clingo.Control(arguments=arguments)
        if observer:
            control.register_observer(observer)
        clingo_session(
            control=control,
            source=source,
            label=label,
            files=files,
            arguments=arguments,
            yield_=True,
            **etc) # not consuming the result so no solving takes place
        return control
    finally:
        os.environ.pop('CLINGOPATH')
        if old_path:
            os.environ['CLINGOPATH'] = ':'.join(old_path)


@test
def ground_simple(stdout):
    control = ground_exc(source="a. b.")
    test.isinstance(control, clingo.Control)
    test.eq(['a', 'b'], list_symbols(control))
    test.endswith(stdout.getvalue(), "-string.lp\n  base()\n")


@test
def ground_files(tmp_path, stdout):
    f1 = write_file(tmp_path/'f1', "f(1).")
    f2 = write_file(tmp_path/'f2', "f(2).")
    control = ground_exc(files=(f1, f2))
    test.eq(['f(1)', 'f(2)'], list_symbols(control))
    out = stdout.getvalue()
    test.contains(out, "/f1\n  base()\n")
    test.contains(out, "/f2\n  base()\n")


@test
def ground_with_observer(stdout):
    atoms = {}
    class MyObserver:
        def output_atom(self, symbol, atom):
            atoms[atom] = symbol
    ground_exc(source="a.", observer=MyObserver())
    test.eq({0: clingo.Function('a', [], True)}, atoms)
    test.endswith(stdout.getvalue(), "-string.lp\n  base()\n")


@test
def pass_arguments_to_control(stdout):
    with test.raises(SyntaxError, "atom does not occur in any rule head:  b"), test.stderr as stderr:
        ground_exc(source="a :- b.")
    test.endswith(stderr.getvalue(), "-string.lp:1:6-7: info: atom does not occur in any rule head:\n  b\n\n")
    c = ground_exc(source="c. a :- b.", arguments=['--warn', 'no-atom-undefined'])
    test.eq(['c'], list_symbols(c))
    test.eq(2, stdout.getvalue().count("-string.lp\n  base()\n"))


@test
def raise_syntaxerror():
    with test.raises(SyntaxError, "syntax error, unexpected EOF") as e:
        ground_exc(source="foutje")
    test.eq('    1 foutje\n      ^ syntax error, unexpected EOF', e.text)


@test
def has_reify(stdout):
    c = ground_exc(source="b. rule(a, b).")
    test.eq(['b', 'a', 'rule(a,b)'], list_symbols(c))
    test.endswith(stdout.getvalue(), "-string.lp\n  base()\n")


@test
def include_path(tmp_path, stdout):
    before = os.environ.get('CLINGOPATH', 'nope')
    (tmp_path/'inc.lp').write_text("included.")
    c = ground_exc(source='#include "inc.lp".', include_paths=[tmp_path.as_posix()])
    test.eq(['included'], list_symbols(c))
    after = os.environ.get('CLINGOPATH', 'nope')
    test.eq(before, after)
    out = stdout.getvalue()
    test.contains(out, "/inc.lp\n  base()\n")
    test.endswith(out, "-string.lp\n  base()\n")


class ContextA:
    def a(self):
        return clingo.String("AA")
    
def context_a_plugin(next, **etc):
    logger, load, _ground, solve = next(**etc)
    def ground(control, parts, context):
        context.add_context(ContextA())
        _ground(control, parts=parts, context=context)
    return logger, load, ground, solve


@test
def use_multiple_contexts():
    class ContextB:
        def b(self):
            return clingo.String("BB")

    aspcode = f"""\
insert_plugin("{__name__}:{context_a_plugin.__qualname__}").
#script (python)
import clingo
def c():
    return clingo.String("CC")
#end.
a(@a()). b(@b()). c(@c()).
"""
    result = session2(
        plugins=(
            source_plugin,
            clingo_control_plugin,
            stdin_to_tempfile_plugin,
            clingo_sequencer_plugin,
            compound_context_plugin,
            insert_plugin_plugin,
            clingo_defaults_plugin
        ),
        source=aspcode,
        context=ContextB(),
        yield_=True)
    test.isinstance(result, clingo.SolveHandle)
    models = 0
    test.eq(
        ['a("AA")', 'b("BB")', 'c("CC")', 'insert_plugin("asp_selftest.integration:context_a_plugin")'],
        [str(a.symbol) for a in result.__control.symbolic_atoms])
    with result as h:
        ###
        ###  get(), resume(), model() etc FUCK UP THE HANDLE !!!!
        ###  They discard the last model and start solving the next one.
        ###  After those call, iteration is BROKEN
        ###
        for m in h:
            models += 1
            test.eq('a("AA") b("BB") c("CC") insert_plugin("asp_selftest.integration:context_a_plugin")', str(m))
    test.eq(1, models)


@test
def without_session_no_problem_with_control():
    def control_plugin(next, source):
        control = clingo.Control()
        def main():
            control.add(source)
            control.ground()
            return control.solve(yield_=True)
        return main
    response = control_plugin(None, source="a. b. c.")
    # reponse saves the control from the GC iff we keep it in a local
    # because the control is in the free variables of response
    test.eq(('control', 'source'), response.__code__.co_freevars)
    # so we call it now, and not in one line as in control_plugin(..)()
    result = response()
    models = 0
    with result:
        for model in result:
            models += 1
            test.eq('a b c', str(model))
    test.eq(1, models)


@test
def maybe_session_is_the_problem():
    def control_plugin(next, source):
        control = clingo.Control()
        def main():
            control.add(source)
            control.ground()
            # we cannot use the trick from previous test because session2() already
            # calls the plugin for us and we loose the control
            # therefor we keep it save on the handle
            # See also clingo_defaults_plugin.
            handle = control.solve(yield_=True)
            handle.__control = control  # save control from GC
            return handle
        return main
    result = session2(plugins=(control_plugin,), source="a. b. c.")  
    models = 0
    with result:
        i = iter(result)
        m = i.__next__()
        models += 1
        m = str(m)
        test.eq('a b c', m)
    test.eq(1, models)





#!/usr/bin/env python3

import coverage
coverage = coverage.Coverage(source_pkgs=["asp_selftest"])
coverage.erase()
coverage.start()

import asp_selftest.plugins.clingo_main_plugin

import asp_selftest.session2
import asp_selftest.integration
import asp_selftest.arguments
import asp_selftest.moretests
import asp_selftest.lib


coverage.stop()
coverage.save()
coverage.html_report()



import sys
import os
import tempfile
import collections
import itertools
import clingo.ast

import selftest
test =  selftest.get_tester(__name__)

from .misc import NA, write_file, format_symbols


def is_testprogram(a):
    if a.ast_type == clingo.ast.ASTType.Program:
        if a.name.startswith('test_'):
            return a.name, [p.name for p in a.parameters]


def gather_tests(files, logger):
    all_tests = collections.defaultdict(dict)

    def _filter_program(ast):
        filename = ast.location.begin.filename
        tests = all_tests[filename]
        if program := is_testprogram(ast):
            name, dependencies = program
            if name in tests:
                raise AssertionError(f"Duplicate test: {name!r} in {filename}.")
            tests[name] = (dependencies, ast.location.begin.line)

    def _logger(code, message):
        if code != clingo.MessageCode.FileIncluded:
            logger(code, message)

    clingo.ast.parse_files(files, callback=_filter_program, logger=_logger)
    return reversed(all_tests.items())


def prepare_test_files(files):
    """Prepare test files, reading from stdin if no files are provided."""
    if not files:
        with open(os.dup(0)) as f:  # read from stdin without removing the data
            asp_code = f.read()
        with tempfile.NamedTemporaryFile('w', suffix='-stdin.lp') as stdin_data:
            print(asp_code, file=stdin_data, flush=True)
            return [stdin_data.name]
    return files


def check_model(model, filename, lineno, fulltestname):
    by_signature = model.context.symbolic_atoms.by_signature
    if failures := [s for s in by_signature('cannot', 1) if model.is_true(s.literal)]: # TODO find test for is_true!
        e = AssertionError(', '.join(str(f.symbol) for f in failures))
        e.add_note(f"File {filename}, line {lineno}, in {fulltestname}. Model follows.")
        symbols = '\n'.join(
                str(s) for s in model.symbols(shown=True)
                if s.name != 'cannot' and not s.name.startswith('_'))
        e.add_note(symbols or '<empty model>')
        raise e

def run_tests(test_info, sub_control, next_plugin):
    testname, (dependencies, lineno) = test_info
    fulltestname = f"{testname}({', '.join(dependencies)})"
    parts = [(testname, [NA for _ in dependencies]), *((d, []) for d in dependencies)]
    print(" ", fulltestname, end='', flush=True)

    sub_logger, sub_load, sub_ground, sub_solve = next_plugin(logger=logger, arguments=new_args, context=context, **etc)
    #sub_control = clingo.Control(arguments=new_args, logger=logger)

    sub_load(sub_control, files=(filename,))
    sub_ground(sub_control, parts=parts, context=context)

    with sub_solve(sub_control, yield_=True) as models:
        for model in models:
            check_model(model, filename, lineno, fulltestname)



def testrunner_plugin(next, run_tests=True, logger=None, arguments=(), context=None, **etc):
    """ Runs all tests in every file separately, during loading. """

    next_logger, _load, ground, solve = next(
        logger=logger,
        arguments=arguments,
        context=context,
        **etc)

    def load(control, files):
        files = prepare_test_files(files)
        for filename, tests in gather_tests(files, logger):
            print("Testing", 'stdin' if filename.endswith('-stdin.lp') else filename)
                        
            for testname, (dependencies, lineno) in [('base', ((), 1)), *tests.items()]:
                new_args=list(itertools.dropwhile(lambda p: not p.startswith('--'), arguments))
                sub_control = clingo.Control(arguments=new_args, logger=logger)
                sub_logger, sub_load, sub_ground, sub_solve = next(logger=logger, arguments=new_args, context=context, **etc)

                fulltestname = f"{testname}({', '.join(dependencies)})"
                print(" ", fulltestname, end='', flush=True)
                try:
                    sub_load(sub_control, files=(filename,))
                    parts = [(testname, [NA for _ in dependencies]), *((d, []) for d in dependencies)]
                    sub_ground(sub_control, parts=parts, context=context)
                    with sub_solve(sub_control, yield_=True) as models:
                        for model in models:
                            check_model(model, filename, lineno, fulltestname)
                finally:
                    print(flush=True)

        _load(control, files)

    return next_logger, load if run_tests else _load, ground, solve


def tracing_clingo_plugin(trace=lambda x: None):
    def tracer(**etc):
        trace(etc)
        def load(control, files):
            trace((control, files))
            control.load(files[0])
        def ground(control, parts, context=None):
            trace((control, parts, context))
            control.ground(parts=parts, context=context)
        def solve(control, yield_):
            trace(yield_)
            return control.solve(yield_=yield_)
        return None, load, ground, solve
    return tracer


@test
def testrunner_plugin_basics(tmp_path, stdout):
    testfile = write_file(tmp_path/'testfile.lp', """\
        #program test_a.
        cannot("I fail").
    """)

    _0, load, ground, solve = testrunner_plugin(tracing_clingo_plugin())

    main_control = clingo.Control()
    with test.raises(AssertionError, 'cannot("I fail")') as e:
        load(main_control, files=(testfile,))
    test.eq(e.exception.__notes__[0], f"File {testfile}, line 1, in test_a(). Model follows.")
    test.eq(stdout.getvalue(), f"Testing {testfile}\n  base()\n  test_a()\n")


@test
def testrunner_plugin_no_failures(tmp_path, stdout):
    testfile = write_file(tmp_path/'testfile.lp', """\
        a.
        #program test_a(base).
        cannot(a) :- not a.
    """)

    _, load, ground, solve = testrunner_plugin(tracing_clingo_plugin())

    main_control = clingo.Control()
    load(main_control, files=(testfile,))
    test.eq(stdout.getvalue(), f"Testing {testfile}\n  base()\n  test_a(base)\n")

    ground(main_control, parts=(('base', ()), ('test_a', ())))
    test.eq('a', str(next(main_control.symbolic_atoms.by_signature('a', 0)).symbol))


@test
def run_tests_per_included_file_separately(tmp_path, stdout, stderr):
    part_a = write_file(tmp_path/'part_a.lp',
        f'#program test_a.')
    part_b = write_file(tmp_path/'part_b.lp',
        f'#include "{part_a}".  #program test_b.')
    part_c = write_file(tmp_path/'part_c.lp',
        f'#include "{part_b}".  #include "{part_a}".  #program test_c.')

    _, load, ground, solve = testrunner_plugin(tracing_clingo_plugin(), arguments=['--warn', 'no-file-included'])
    main_control = clingo.Control(arguments=['--warn', 'no-file-included'])
    load(main_control, files=(part_c,))
    test.eq('', stderr.getvalue())
    out = stdout.getvalue()
    test.contains(out, f"""\
Testing {part_b}
  base()
  test_b()
Testing {part_a}
  base()
  test_a()
Testing {part_c}
  base()
  test_c()""")


# TEST taken from runasptests.py


def parse_and_run_tests(asp_code, trace=lambda _:None, **etc):
    with test.tmp_path as p:
        inputfile = write_file(p/'inputfile.lp', asp_code)
        _, load, _, _ = testrunner_plugin(
            tracing_clingo_plugin(trace=trace), **etc)
        main_control = clingo.Control()
        load(main_control, files=(inputfile,))
    

@test
def cannot_in_base(stdout):
    with test.raises(AssertionError, 'cannot(fail)'):
        parse_and_run_tests("cannot(fail).")  # constraints in base
    test.endswith(stdout.getvalue(), "/inputfile.lp\n  base()\n")
    

@test
def use_arguments_for_testing(stdout):
    trace = []
    asp_program = "sum(a). cannot(a) :- not sum(42)."
    parse_and_run_tests(
        asp_program,
        arguments=['--const', 'a=42'],
        trace=trace.append)
    test.eq({'logger': None, 'arguments': ['--const', 'a=42'], 'context': None}, trace[0])
    with test.raises(AssertionError, "cannot(99)"):
        parse_and_run_tests(
            asp_program,
            arguments=['--const', 'a=99'])
    test.endswith(stdout.getvalue(), "/inputfile.lp\n  base()\n")



@test
def use_context_when_running_tests(stdout):
    trace = []
    asp_program = "#program test_a_42. a(42). cannot(@a()) :- a(@a())."
    n = 43
    class MyContext:
        def a(self):
            return clingo.Number(n)
    mycontext = MyContext()
    parse_and_run_tests(
        asp_program,
        context=mycontext,
        trace=trace.append)
    test.eq(mycontext, trace[0]['context'])
    n = 42
    with test.raises(AssertionError, "cannot(42)"):
        parse_and_run_tests(
            asp_program,
            context=MyContext())
    test.endswith(stdout.getvalue(), "/inputfile.lp\n  base()\n  test_a_42()\n")
    
        
@test
def check_for_duplicate_test():
    with test.raises(Exception) as e:
        parse_and_run_tests(""" #program test_a. \n #program test_a. """)
    test.startswith(str(e.exception), "Duplicate test: 'test_a' in ")


@test
def format_empty_model(stderr, stdout):
    with test.raises(AssertionError) as e:
        parse_and_run_tests("""
            #program test_model_formatting.
            #external what.
            cannot(test) :- not what.
        """)
    msg = str(e.exception)
    #test.eq(msg, f"""MODEL:
    #<empty>
    #Failures in <string>, #program test_model_formatting():
    #cannot(test)
    #""")
    notes = e.exception.__notes__
    test.startswith(notes[0], "File ")
    test.endswith(notes[0], ", line 2, in test_model_formatting(). Model follows.")
    test.eq(notes[1], "<empty model>")


@test
def format_model_small(stderr, stdout):
    with test.raises(AssertionError) as e:
        parse_and_run_tests("""
            #program test_model_formatting.
            this_is_a_fact(1..2).
            #external what.
            cannot(test) :- not what.
        """)
    notes = e.exception.__notes__
    test.startswith(notes[0], "File ")
    test.endswith(notes[0], ", line 2, in test_model_formatting(). Model follows.")
    test.eq(notes[1], """this_is_a_fact(1)\nthis_is_a_fact(2)""")


@test
def format_model_wide(stderr, stdout):
    with test.raises(AssertionError) as e:
        parse_and_run_tests("""
            #program test_model_formatting.
            this_is_a_fact(1..3).
            #external what.
            cannot(test) :- not what.
        """)
    notes = e.exception.__notes__
    test.startswith(notes[0], "File ")
    test.endswith(notes[0], ", line 2, in test_model_formatting(). Model follows.")
    test.eq(notes[1], """this_is_a_fact(1)\nthis_is_a_fact(2)\nthis_is_a_fact(3)""")


@test
def we_CAN_NOT_i_repeat_NOT_reuse_control():
    c = clingo.Control()
    c.add("a. #program p1. p(1). #program p2. p(2).")
    c.ground()
    test.eq(['a'], [str(s.symbol) for s in c.symbolic_atoms])
    c.cleanup()
    c.ground((('base', ()), ('p1', ())))
    test.eq(['a', 'p(1)'], [str(s.symbol) for s in c.symbolic_atoms])
    c.cleanup()
    c.ground((('base', ()), ('p2', ())))
    # p(1) should be gone
    test.eq(['a', 'p(1)', 'p(2)'], [str(s.symbol) for s in c.symbolic_atoms])


@test
def dependencies(stderr, stdout):
    parse_and_run_tests("""
        base_fact.

        #program one().
        one_fact.

        #program test_base(base).
        cannot(base_fact) :- not base_fact.

        #program test_one(base, one).
        cannot(base_fact) :- not base_fact.
        cannot(one_fact) :- not one_fact.
    """)
    test.eq('', stderr.getvalue())
    test.endswith(stdout.getvalue(),
      "/inputfile.lp\n  base()\n  test_base(base)\n  test_one(base, one)\n")


@test
def run_tests_flag(stdout):
    code = "#program test_a. cannot(true)."
    with test.raises(AssertionError, "cannot(true)"):
        parse_and_run_tests(code, run_tests=True)
    with test.stdout as out:
        parse_and_run_tests(code, run_tests=False)
    test.eq('', out.getvalue())




import collections
import clingo
import clingo.ast
import contextlib

import selftest
test = selftest.get_tester(__name__)


""" This module contains ASP-specific utilities for dealing with various
    clingo data representations of similar concepts such as Symbols, 
    Functions, String etc.
    The same functions work for (but check!) for multiple representations:

     - ground symbols (clingo.symbol.Symbol etc)      (most)
     - parser symbols (clingo.ast.SymbolicAtom etc)   (many)
     - theory symbols (clingo.theory.TheoryAtom etc)  (few)
 """

A_THEORY = '#theory a { t{}; &p/1: t, head}.' # just a theory for testing


def is_ast(obj):
    return isinstance(obj, clingo.ast.AST)


class Literal(collections.namedtuple('Literal', ['sign', 'symbol', 'type'])):
    def __str__(self):
        return f"not {self.symbol}" if self.sign else str(self.symbol)


def is_literal(obj):
    if is_ast(obj):
        return obj.ast_type == clingo.ast.ASTType.Literal
    if isinstance(obj, Literal):
        return True


def is_function(obj, name=None):
    if is_ast(obj):
        return obj.ast_type == clingo.ast.ASTType.Function and (name is None or obj.name == name)
    if is_literal(obj):
        return obj.type == clingo.SymbolType.Function and (name is None or obj.symbol.name == name)
    if is_theory_term(obj):
        assert name is None
        return obj.type == clingo.TheoryTermType.Function
    return obj.type == clingo.SymbolType.Function and (name is None or obj.name == name)


def is_tuple(s):
    if isinstance(s, clingo.TheoryTerm):
        return s.type == clingo.TheoryTermType.Tuple
    return is_function(s, '')


def is_number(symbol):
    return symbol.type == clingo.SymbolType.Number


def is_string(symbol):
    if isinstance(symbol, clingo.TheoryTerm):
        s = symbol
        return s.type == clingo.TheoryTermType.Symbol and s.name[0] == s.name[-1] == '"'
    return symbol.type == clingo.SymbolType.String


def is_symbol(symbol):
    return symbol.type == clingo.TheoryTermType.Symbol


def is_variable(ast):
    return ast.ast_type == clingo.ast.ASTType.Variable


def is_symbolicterm(ast):
    return ast.ast_type == clingo.ast.ASTType.SymbolicTerm


def is_conditional(ast):
    return ast.ast_type == clingo.ast.ASTType.ConditionalLiteral


def is_disjunction(ast):
    return ast.ast_type == clingo.ast.ASTType.Disjunction


def is_positive(ast):
    if is_ast(ast):
        lit = ast.literal if is_conditional(ast) else ast
        return lit.sign == 0
    return not ast.sign


def is_binaryoperation(ast):
    return ast.ast_type == clingo.ast.ASTType.BinaryOperation


def is_minusoperation(ast):
    return is_unaryoperation(ast) and ast.operator_type == clingo.ast.UnaryOperator.Minus


def is_unaryoperation(ast):
    return ast.ast_type == clingo.ast.ASTType.UnaryOperation


def mk_ast(lit):
    if is_ast(lit):
        return lit
    source_rules = []
    def cb(ast):
        if ast.ast_type == clingo.ast.ASTType.Rule:
            source_rules.append((ast.head, ast.body))
    clingo.ast.parse_string(lit+'.', cb)
    return source_rules[0][0]


mk_symbol = clingo.parse_term


def mk_literal(literal):
    sign = literal.startswith('not')
    symbol = mk_symbol(literal[4 if sign else 0:])
    return Literal(sign=sign, symbol=symbol, type=symbol.type)


@contextlib.contextmanager
def mk_theory_atom(asp, theory):
    c = clingo.Control()
    c.add(theory)
    c.add(asp)
    c.ground()
    yield next(c.theory_atoms)


def is_theory_term(symbol):
    return isinstance(symbol, clingo.TheoryTerm)


@test
def simple_terms():
    time = mk_literal('time')
    test.eq("time", time.symbol.name)
    test.eq(clingo.SymbolType.Function, time.symbol.type)
    test.eq('time', str(time))
    test.truth(is_function(time))
    test.truth(is_function(time, name='time'))
    test.not_ (is_function(time, name='fire'))
    test.truth(is_function(mk_literal('time(3)')))
    test.truth(is_function(mk_literal('time(3)'), name='time'))


@test
def simple_numbers():
    time = mk_literal('23')
    test.eq(clingo.SymbolType.Number, time.type)
    test.eq(23, time.symbol.number)
    test.truth(is_number(time))
    test.not_ (is_function(time))
    test.not_ (is_string(time))


@test
def simple_strings():
    time = mk_literal('"24"')
    test.eq(clingo.SymbolType.String, time.type)
    test.eq("24", time.symbol.string)
    test.truth(is_string(time))
    test.not_ (is_function(time))
    test.not_ (is_number(time))


@test
def is_string_theory():
    with mk_theory_atom('&p("aap").', A_THEORY) as ta:
        test.truth(is_string(ta.term.arguments[0]))



@test
def function_wit_arguments():
    time = mk_literal('time(24, "H")')
    test.eq('time(24,"H")', str(time))
    test.truth(is_function(time))
    test.not_ (time.sign)
    test.eq("time", time.symbol.name)
    test.truth(is_function(time))
    args = time.symbol.arguments
    test.truth(is_number(args[0]))
    test.truth(is_string(args[1]))


@test
def some_asts_and_their_structure():
    test.truth(is_literal(mk_ast('bool')))
    test.truth(is_literal(mk_ast('bool(1)')))
    test.truth(is_literal(mk_ast('bool(a, 1)')))
    l = mk_ast('not bool(a, A, 1, "A", f(a))')
    l2 = mk_ast(l)
    test.eq(l, l2)
    test.truth(is_literal(l))
    b = l.atom.symbol
    test.truth(is_function(b, name='bool'))
    test.not_ (is_function(b, name='loob'))
    a, A, one, strA, fa = b.arguments
    test.truth(is_symbolicterm(a))
    test.truth(is_function(a.symbol))
    test.eq('a', a.symbol.name)
    test.truth(is_variable(A))
    test.eq('A', A.name)
    test.truth(is_symbolicterm(one))
    test.truth(is_number(one.symbol))
    test.truth(is_symbolicterm(strA))
    test.truth(is_string(strA.symbol))
    test.truth(is_function(fa))


@test
def conditional_positive_literals():
    d0 = mk_ast('b : c')
    test.truth(is_disjunction(d0))
    test.not_ (is_conditional(d0))
    c0 = d0.elements[0]
    test.truth(is_conditional(c0))
    test.not_ (is_disjunction(c0))
    test.truth(is_positive(c0))
    l0 = c0.literal
    test.truth(is_literal(l0))
    test.not_ (is_conditional(l0))
    test.not_ (is_disjunction(l0))
    test.truth(is_positive(l0))


@test
def conditional_negative_literals():
    d1 = mk_ast('not b : c')
    test.truth(is_disjunction(d1))
    test.not_ (is_conditional(d1))
    c1 = d1.elements[0]
    test.truth(is_conditional(c1))
    test.not_ (is_disjunction(c1))
    test.not_ (is_positive(c1))
    l1 = c1.literal
    test.truth(is_literal(l1))
    test.not_ (is_conditional(l1))
    test.not_ (is_disjunction(l1))
    test.not_ (is_positive(l1))


@test
def pos_and_neg_literal_symbol():
    posA = mk_literal('a')
    test.truth(is_literal(posA))
    test.truth(is_positive(posA))
    negA = mk_literal('not a')
    test.eq('not a', str(negA))
    test.truth(is_literal(negA))
    test.not_(is_positive(negA))


@test
def symbolic_atoms():
    l = mk_literal('not b(0,e)')
    test.eq(True, l.sign)
    test.eq(mk_literal('b(0,e)').symbol, l.symbol)


@test
def unary_op_and_minus():
    u = mk_ast("f(-A)").atom.symbol.arguments[0]
    test.truth(is_unaryoperation(u))
    test.eq(clingo.ast.UnaryOperator.Minus, u.operator_type)
    test.truth(is_minusoperation(u))
    u = mk_ast("f(|A|)").atom.symbol.arguments[0]
    test.truth(is_unaryoperation(u))
    test.eq(clingo.ast.UnaryOperator.Absolute, u.operator_type)
    test.not_(is_minusoperation(u))
    u = mk_ast("f(~A)").atom.symbol.arguments[0]
    test.truth(is_unaryoperation(u))
    test.eq(clingo.ast.UnaryOperator.Negation, u.operator_type)
    test.not_(is_minusoperation(u))


@test
def bin_op():
    b = mk_ast("f(1+2)").atom.symbol.arguments[0]
    test.truth(is_binaryoperation(b))


@test
def is_tuple_with_theory():
    test.truth(is_tuple(mk_literal('(1,2,3)')))
    with mk_theory_atom('&p((1,2,3)).', A_THEORY) as ta:
        test.truth(is_tuple(ta.term.arguments[0]))




"""

    EXPERIMENT with easier way to use plugins.

    It tries to separate the Clingo-specific sequencing:
        Control() -> Load -> Ground -> Solve

    from the plugin logic. Or: make the sequencing a plugin.

"""

import selftest
test = selftest.get_tester(__name__)

from .plugins.misc import write_file, ExitCode

from .plugins import (
    clingo_main_plugin,
    source_plugin,
    clingo_control_plugin,
    clingo_syntaxerror_plugin,
    clingo_sequencer_plugin,
    clingo_defaults_plugin,
    testrunner_plugin,
    clingo_reify_plugin,
    stdin_to_tempfile_plugin,
)

import clingo
VERSION = '.'.join(map(str,clingo.version()))


def session2(plugins=(), **etc):
    """ Calls each plugin (factory) with the next one as first argument, followed by **etc.
        The first plugin must return a callable wich is called immediately. """
    assert len(plugins) > 0, plugins
    def get_plugin_func(i):
        def get_plugin(**etc):
            assert i < len(plugins), f"No more plugins after '{plugins[-1].__name__}'"
            return plugins[i](get_plugin_func(i+1), **etc)
        return get_plugin
    return get_plugin_func(0)(**etc)()


@test
def test_session2():
    def hello_plugin(next, name=None):
        """ A plugin is called with the next plugin as first argument, followed by all the given keyword args.
            The first plugin must return a callable. """
        with test.raises(AssertionError, "No more plugins after 'hello_plugin'"):
            next()
        def hello():
            return f"Hello {name}"
        return hello
    test.eq("Hello John", session2(plugins=(hello_plugin,), name="John"))


@test
def test_session2_sequencing():
    trace = []
    def hello_goodbye_plugin(next, name=None):
        """ This plugin only works with the sequencer, which expects two functions."""
        def hi():
            trace.append(f"Hi {name}!")
        def jo():
            trace.append(f"Jo {name}!")
        return hi, jo
    def sequencer_plugin(next, **etc):
        """ This plugin expects two functions and returns one (as required)."""
        hello, goodbye = next(**etc)
        def main():
            hello()
            goodbye()
        return main
    test.eq(None, session2(plugins=(sequencer_plugin, hello_goodbye_plugin,), name="John"))
    test.eq(['Hi John!', 'Jo John!'], trace)

   
common_plugins = (
    clingo_syntaxerror_plugin,
    clingo_sequencer_plugin,
    testrunner_plugin,
    clingo_reify_plugin,
    clingo_defaults_plugin,
)

def clingo_main_session(**kwargs):
    return session2(
        plugins=(
            clingo_main_plugin,
            stdin_to_tempfile_plugin,
            *common_plugins),
        **kwargs)

def clingo_session(**kwargs):
    return session2(
        plugins=(
            source_plugin,
            clingo_control_plugin,
            *common_plugins),
        **kwargs)


@test
def clingo_main_session_happy_flow(stdout, tmp_path):
    file1 = tmp_path/'file1.lp'
    file1.write_text('a.')
    exitcode = clingo_main_session(arguments=(file1.as_posix(),))
    test.eq(exitcode, ExitCode.SAT)
    out = stdout.getvalue()
    test.startswith(out, f"clingo+ version {VERSION}\nReading from ")
    i = out.find("Time: 0.0")
    j = out.find("s)\na\n")
    out = out[:i] + 'xxx' + out[j:] 
    test.contains(out, "Answer: 1 (xxxs)\na\nSATISFIABLE\n\nModels       : 1+\n")


@test
def clingo_main_session_error(tmp_path, stdout, stderr):
    file1 = write_file(tmp_path/'error.lp', 'error')
    with test.raises(SyntaxError, "syntax error, unexpected EOF") as e:
        clingo_main_session(arguments=(file1,))
    err = stderr.getvalue()
    out = stdout.getvalue()
    test.eq('', err)
    test.startswith(out, f"clingo+ version {VERSION}\nReading from ")
    test.contains(out, f"{file1[-38:]}\nUNKNOWN\n\nModels       : 0+\nCalls        : 1\nTime")
    test.endswith(e.exception.text, "    1 error\n      ^ syntax error, unexpected EOF")


@test
def session_with_source(stderr):
    with test.raises(SyntaxError) as e:
        clingo_session(source="error", label='yellow')
    msg = str(e.exception)
    test.startswith(msg, "syntax error, unexpected EOF (tmp")
    test.endswith(msg, f"-yellow.lp, line 2)")


@test
def test_session2_not_wat(stdout):
    solve_handle = clingo_session(source="a. b. c(a).", yield_=True)
    with solve_handle as result:
        for model in result:
            test.eq('a b c(a)', str(model))
    test.endswith(stdout.getvalue(), "-string.lp\n  base()\n")


@test
def session_with_file(tmp_path, stdout):
    file1 = tmp_path/'test.lp'
    file1.write_text('test(1).')
    solveresult = clingo_session(files=(file1.as_posix(),))
    test.truth(solveresult.satisfiable)
    test.endswith(stdout.getvalue(), "test.lp\n  base()\n")




""" Runs all tests in an ASP program.

    This module contains all 'mains', e.q. entry points as 
    defined in pyproject.toml.

    Tests are in moretests.py to keep to module importable with choice of running tests or not
"""

import sys


# this function is directly executed by pip installed code wrapper, see pyproject.toml
def clingo_plus():

    from .arguments import maybe_silence_tester
    remaining = maybe_silence_tester()

    from .arguments import parse_plus_arguments
    args, remaining = parse_plus_arguments(remaining)

    from .session2 import clingo_main_session
    clingo_main_session(run_tests=args.run_asp_tests, arguments=remaining)
    
    #import cProfile
    #with cProfile.Profile() as p:
    #    try:
    #        main_clingo_plus(clingo_options, programs=plus_options.programs)
    #    finally:
    #        p.dump_stats('profile.prof') # use snakeviz to view



import sys

import selftest
test = selftest.get_tester(__name__)


class CompoundContext:
    """ Clingo looks up functions in __main__ OR in context; we need both.
        (Functions defined in #script land in __main__)
    """

    def __init__(self, *contexts):
        self._contexts = list(contexts)


    def add_context(self, *context):
        self._contexts += context
        return self


    def __getattr__(self, name):
        for c in self._contexts:
            if f := getattr(c, name, None):
                return f
        return getattr(sys.modules['__main__'], name)


def compound_context_plugin(next, **etc):

    logger, load, _ground, solve = next(**etc)

    def ground(control, parts=(('base', ()),), context=None):
        compound_context = CompoundContext()
        if context:
            compound_context.add_context(context)
        _ground(control, parts=parts, context=compound_context)

    return logger, load, ground, solve
    
    
# tests in integration.py



import clingo
from .misc import write_file, ExitCode

import selftest
test = selftest.get_tester(__name__)


VERSION = '.'.join(map(str,clingo.version()))

def clingo_main_plugin(next, arguments=(), **etc):
    """ Uses clingo_main() to drive the plugins. It is meant for implementing a Clingo compatible
        main. It does not return anything as to avoid garbage collected C++ objects to ruin the program.
    """
            
    class App: #(clingo.Application):
        """ As per Clingo spec: callbacks main() and logger(). """
        program_name = 'clingo+'
        exceptions = []
                
        def main(self, control, files):
            """ As required by clingo_main. It must not raise. """
            try:
                self._logger, _main = next(logger=self.logger, control=control, files=files, arguments=arguments, **etc)  # [3]
                return _main()   #  [4]
            except Exception as e:
                self.exceptions.append(e)
                    
        def logger(self, code, message):
            """ As required by clingo_main. Forwards to next plugin."""
            try:
                self._logger(code, message)
            except Exception as e:
                self.exceptions.append(e)

    def main():
        app = App()
        exitcode = clingo.clingo_main(app, arguments)  # [2]
        if app.exceptions:
            raise app.exceptions[0]  # we report only the first one
        return exitcode
            
    return main  #[1]


@test
def clingo_thingies(stdout):
    def next_plugin():
        return None, lambda: None
    clingo_main_plugin(next_plugin, arguments=["--help"])()
    out = stdout.getvalue()
    test.startswith(out, "clingo+ version 5.8.0\nusage: clingo+ [number] [options] [files]\n\n")


@test
def raise_errors_in_plugins(stdout):
    def malicious_plugin(**etc):
        return 1, 2, 3
    main = clingo_main_plugin(malicious_plugin, arguments=[])
    with test.raises(ValueError, "too many values to unpack (expected 2)"):
        main()
    test.startswith(stdout.getvalue(), f"""clingo+ version {VERSION}
Reading from stdin
UNKNOWN""")


@test
def raise_errors_in_main(stdout):
    def malicious_plugin(**etc):
        def malicious_main():
            1/0
        return None, malicious_main
    main = clingo_main_plugin(malicious_plugin, arguments=[])
    with test.raises(ZeroDivisionError, "division by zero"):
        main()
    test.startswith(stdout.getvalue(), f"""clingo+ version {VERSION}
Reading from stdin
UNKNOWN""")


@test
def raise_exception_in_logger(stdout):

    def plugin_raising_in_logger(control=None, **__):
        def logger(code, message):
            raise TypeError("oh no!")
        def main():
            control.add("not good")
        return logger, main

    main = clingo_main_plugin(plugin_raising_in_logger)
    with test.raises(TypeError, "oh no!"):
        main()
    test.startswith(stdout.getvalue(), "clingo+ version 5.8.0")


@test
def plugin_basic_noop(stdout):
    arguments = []
    def next(*args, **etc):
        # !! we test only the plugin; hance we have no 'next' as first argument
        #    when called with session(), we have the next plugin as 'next'.
        arguments.append(args)
        arguments.append(etc)
        return lambda *a: arguments.append(a), next
    main = clingo_main_plugin(next, arguments=[], etc='42')
    test.eq([], arguments)
    exitcode = main()
    test.eq(ExitCode.UNKNOWN, exitcode)
    test.eq(4, len(arguments))
    plugin_args = arguments[0]
    test.eq((), plugin_args)
    plugin_kwargs = arguments[1]
    test.isinstance(plugin_kwargs['control'], clingo.Control)
    test.eq([], plugin_kwargs['files'])
    test.eq('42', plugin_kwargs['etc'])
    test.eq([], plugin_kwargs['arguments'])
    logger = plugin_kwargs['logger']
    test.ismethod(logger)
    test.eq(5, len(plugin_kwargs))
    main_args = arguments[2:5]
    test.eq([(), {}], main_args)
    logger(42, "he!")  # should call our own logger
    test.eq((42, 'he!'), arguments[-1])


@test
def pass_arguments_to_files(tmp_path, stdout):
    f1 = write_file(tmp_path/"f1.lp", "f1.")
    f2 = write_file(tmp_path/"f2.lp", "f2.")
    trace = []
    def next_main():
        trace.append('main')
    def next(files=None, **etc):
        trace.append(files)
        trace.append(etc)
        return None, next_main
    main = clingo_main_plugin(next, arguments=[f1, f2])
    main()
    test.eq([f1, f2], trace[0])
    test.isinstance(trace[1]['control'], clingo.Control)
    test.eq('main', trace[2])
    test.eq(3, len(trace))


@test
def forward_logger(stdout):
    trace = []
    def next(logger=None, control=None, files=(), arguments=()):
        def next_main():
            control.add("error")  # trigger call of logger
        def logger(code, message):
            trace.append(message)
        return logger, next_main
    main = clingo_main_plugin(next)
    with test.raises(RuntimeError, "parsing failed"):
        main()
    test.eq('<block>:2:1-2: error: syntax error, unexpected EOF\n', trace[0])


@test
def pass_arguments_to_next_plugin(stdout):
    trace = []
    def next_plugin(logger=None, control=None, files=(), arguments=()):
        trace[:] = arguments
        return 1, lambda: None
    main = clingo_main_plugin(next_plugin, arguments=['--project=no'])
    test.eq([], trace)
    exitcode = main()
    test.eq(0, exitcode)
    test.eq(['--project=no'], trace)



import functools
import importlib
import clingo
from .misc import write_file, is_plugin_instruction

import selftest
test = selftest.get_tester(__name__)


def insert_plugin_plugin(next, **etc):
    next_plugin = next

    def install_plugins(files):
        
        def get_inserts(ast):
            nonlocal next_plugin
            if plugin_name := is_plugin_instruction(ast):
                if ':' in plugin_name:
                    modulename, functionname = plugin_name.rsplit(':', 1)
                else:
                    modulename, functionname = __name__, plugin_name
                module = importlib.import_module(modulename)
                plugin_function = getattr(module, functionname)
                next_plugin = functools.partial(plugin_function, next_plugin)
   
        clingo.ast.parse_files(files, callback=get_inserts, logger=logger)

        nonlocal _logger, _load, _ground, _solve
        _logger, _load, _ground, _solve = next_plugin(**etc)

    _logger = _load = _ground = _solve = None

    def logger(*a):
        nonlocal _logger
        if _logger is None:  # plugin not called yet, do it now
            _logger, _, _, _ = next_plugin(**etc)
        _logger(*a)

    def load(control, files, *a, **k):
        install_plugins(files)
        _load(control, files, *a, **k)

    # we just defer the lookup of the functions until the plugin is called
    def ground(*a, **k):
        _ground(*a, **k)
        
    def solve(*a, **k):
        return _solve(*a, **k)

    return logger, load, ground, solve

def marker_plugin(next, **etc):
    logger, _load, ground, solve = next(**etc)
    def load(control, files):
        _load(control, files)
        control.add("marker.")
    return logger, load, ground, solve

@test
def insert_plugin_basics(tmp_path):
    _trace = []
    trace = _trace.append
    f = write_file(tmp_path/'f.lp', 'insert_plugin("marker_plugin"). main.')
    def next_plugin(**etc): # no next plugin here b/c testing
        def logger(code, message):
            trace(f"logger: {code}, {message}")
        def load(control, files):
            trace(f"load: {files}")
            control.load(files[0])
        def ground(control, parts=(('base', ()),), context=None):
            trace(f"ground: {parts}, {context}")
            control.ground(parts=parts, context=context)
        def solve(control):
            return "next-plugin"
        return logger, load, ground, solve
    logger, load, ground, solve = insert_plugin_plugin(next_plugin)
    logger(42, "help!")
    test.eq("logger: 42, help!", _trace[0])
    control = clingo.Control()
    load(control, (f,))
    ground(control)
    test.eq(
        {'insert_plugin("marker_plugin")', 'main', 'marker'},
        {str(a.symbol) for a in control.symbolic_atoms})
    test.eq(f"load: ({f!r},)", _trace[1])
    test.eq("ground: (('base', ()),), None", _trace[2])
    test.eq(3, len(_trace))
    test.eq('next-plugin', solve(None))


def look_plugin(next, **etc):
    _logger, *rest = next(**etc)
    def logger(code, message):
        _logger(code, "LOOK! " + message)
    return logger, *rest


@test
def fall_back_to_latest_plugin(tmp_path):
    # force parse error; this happens during parsing but the new plugin is
    # added already, so it must use it
    f = write_file(tmp_path/'f.lp', """\
        insert_plugin("look_plugin").
        parse error.""")
    trace = []
    def next_plugin(**etc): # no next plugin here b/c testing
        def logger(code, message):
            trace.append((code.value, message))
        return logger, None, None, None
    logger, load, ground, solve = insert_plugin_plugin(next_plugin)
    with test.raises(RuntimeError, "syntax error"):
        load(None, (f,))
    test.eq((1, f"LOOK! {f}:2:15-20: error: syntax error, unexpected <IDENTIFIER>\n"), trace[0])


@test
def use_proper_logger(tmp_path):
    # an parse error occurs before the plugin is inserted
    f = write_file(tmp_path/'f.lp', """\
        parse error.
        insert_plugin("look_plugin").""")
    trace = []
    def next_plugin(**etc): # no next plugin here b/c testing
        def logger(code, message):
            trace.append((code.value, message))
        return logger, None, None, None
    logger, load, ground, solve = insert_plugin_plugin(next_plugin)
    with test.raises(RuntimeError, "syntax error"):
        load(None, (f,))
    test.eq((1, f"{f}:1:15-20: error: syntax error, unexpected <IDENTIFIER>\n"), trace[0])


#TODO test error conditions (maybe if you really start using this idea)




import sys
import clingo

from .misc import write_file

import selftest
test = selftest.get_tester(__name__)


def clingo_defaults_plugin(next, **etc):
    """ Implements Clingo sequence with default actions. """
    
    def logger(code, message):
        print(f"UNHANDLED MESSAGE: code={code}, message: {message!r}", file=sys.stderr)
                
    def load(control, files=()):
        for filename in files:
            control.load(filename)
        if not files:
            control.load('-')

    def ground(control, **kw):
        control.ground(**kw)

    def solve(control, **kw):
        result = control.solve(**kw)
        # Clingo Python/C++ API does not maintain a relation between a Handle and its Control.
        # As soon as the control goes out of scope, the Handle can no longer work.
        # Therefor we keep the control save by poking it on the handle ourselves.
        result.__control = control # save control from GC.
        return result
                    
    return logger, load, ground, solve
    

@test
def clingo_defaults_plugin_basics(tmp_path, stderr):
    file1 = write_file(tmp_path/'file1.lp', 'a. b.')
    control = clingo.Control()
    _, load, ground, solve = clingo_defaults_plugin(None)
    load(control, files=(file1,))
    ground(control)
    test.eq('a', str(next(control.symbolic_atoms.by_signature('a', 0)).symbol))
    test.eq('b', str(next(control.symbolic_atoms.by_signature('b', 0)).symbol))
    models = []
    solve(control, on_model=lambda model: models.append(str(model)))
    test.eq(['a b'], models)


@test
def clingo_defaults_plugin_logger(stderr):
    control = clingo.Control()
    logger, l, g, s = clingo_defaults_plugin(None)
    logger(67, 'message in a bottle')
    test.eq(stderr.getvalue(), "UNHANDLED MESSAGE: code=67, message: 'message in a bottle'\n")


@test
def keep_control_from_GC():
    _, l, g, s = clingo_defaults_plugin(None)
    control = clingo.Control()
    result = s(control)
    test.eq(control, result.__control)


#@test
def no_files():
    import os
    logger, load, ground, solve = clingo_defaults_plugin(None)
    control = clingo.Control()
    os.write(0, b"aap")
    load(control, files=())
    ground(control)
    with solve(control, yield_=True) as s:
        for m in s:
            print(m)




""" Program argument and in-source test handling + some tests that introduce cyclic dependencies elsewhere. """

import os
import sys
import subprocess
import pathlib


import clingo
from clingo import Control


from .arguments import maybe_silence_tester
from .__main__ import clingo_plus


import selftest
test = selftest.get_tester(__name__)


VERSION = '.'.join(map(str,clingo.version()))


def spawn_clingo_plus(input="", arguments=[]):
    path = pathlib.Path(__file__).parent
    p = subprocess.run(["python", "-c", f"from asp_selftest.__main__ import clingo_plus; clingo_plus()"] + arguments,
        env=os.environ | {'PYTHONPATH': path},
        input=input,
        capture_output=True)
    return p


@test
def maybe_shutup_selftest(argv):
    argv += ['--silent']
    try:
        maybe_silence_tester()
    except AssertionError as e:
        test.startswith(str(e), 'In order to NOT run Python tests, Tester <Tester None created at:')
        test.endswith(str(e), 'must have been configured to NOT run tests.')
    # this indirectly tests if the code above actually threw the AssertionError
    test.eq(True, selftest.get_tester(None).option_get('run'))


@test
def main_entry_point_basics():
    p = spawn_clingo_plus(input=b"skaludicat. #program test_gotashparot(base).", arguments=['--run-asp-tests'])
    test.eq(b'', p.stderr)
    test.startswith(p.stdout.decode(), f"""\
clingo+ version {VERSION}
Reading from stdin
Testing stdin
  base()
  test_gotashparot(base)
Solving...
Answer: 1 (Time: 0.001s)
skaludicat
SATISFIABLE

Models""", diff=test.diff)


@test
def simple_syntax_error_with_clingo_main():
    p = spawn_clingo_plus(b'plugin(".:errorplugin"). a', arguments=['--run-asp-tests'])
    test.startswith(p.stdout, f'clingo+ version {VERSION}\nReading from stdin\nUNKNOWN'.encode())
    traceback = p.stderr
    should = b"""-stdin.lp", line 2
    1 plugin(".:errorplugin"). a
      ^ syntax error, unexpected EOF
asp_selftest.plugins.messageparser.AspSyntaxError: syntax error, unexpected EOF
"""
    test.endswith(traceback, should, diff=test.diff)
    test.comp.contains(p.stdout, b"*** ERROR")
    test.comp.contains(traceback, b"*** ERROR")


@test
def clingo_drop_in_plus_tests(tmp_path, argv, stdout, stderr):
    f = tmp_path/'f.lp'
    f.write_text('a. #program test_ikel(base).\n')
    argv += [f.as_posix(), '--run-python-tests']  # we can not NOT run the Python tests here
    clingo_plus()
    s = iter(stdout.getvalue().splitlines())
    test.eq(f'clingo+ version {VERSION}', next(s))
    l = next(s)
    test.startswith(l, 'Reading from')
    test.endswith(l, 'f.lp')
    test.eq('Solving...', next(s))
    test.eq('Answer: 1 (Time: 0.000s)', next(s))
    test.eq('a', next(s))
    test.eq('SATISFIABLE', next(s))
    test.eq('', next(s))
    test.eq('Models       : 1+', next(s))
    test.eq('Calls        : 1', next(s))
    l = next(s)
    test.contains(l, 'Time')
    test.contains(l, 'Solving:')
    test.contains(l, '1st Model:')
    test.contains(l, 'Unsat:')
    test.startswith(next(s), 'CPU Time     : 0.00')
    test.startswith(stderr.getvalue(), "")


@test
def syntax_errors_basics(tmp_path, argv, stdout, stderr):
    f = tmp_path/'f'
    f.write_text("a syntax error")
    argv += [f.as_posix(), '--run-python-tests']
    with test.raises(SyntaxError) as e:
        clingo_plus()
    out = stdout.getvalue()
    test.eq('syntax error, unexpected <IDENTIFIER>', e.exception.msg)
    test.contains(out, f"Reading from ...{f.as_posix()[-38:]}")
    err = stderr.getvalue()
    test.eq(err, '')


@test
def tester_runs_tests(tmp_path, argv, stdout, stderr):
    f = tmp_path/'f'
    f.write_text("""
    fact(a).
    #program test_fact(base).
    cannot("fact") :- not fact(a).
    models(1).
    """)
    argv += [f.as_posix(), '--run-asp-tests', '--run-python-tests']
    clingo_plus()
    test.contains(stdout.getvalue(), f"Testing {f}\n  base()\n  test_fact(base)\n")
    test.startswith(stderr.getvalue(), "")


@test
def clingo_dropin_default_hook_tests(tmp_path, argv, stdout, stderr):
    f = tmp_path/'f'
    f.write_text("""
    fact(a).
    #program test_fact_1(base).
    cannot("fact 1") :- not fact(a).
    models(1).
    #program test_fact_2(base).
    cannot("fact 2") :- not fact(a).
    models(1).
    """)
    argv += [f.as_posix(), '--run-asp-tests', '--run-python-tests']
    clingo_plus()
    s = stdout.getvalue()
    test.contains(s, f"Testing {f}\n  base()\n  test_fact_1(base)\n  test_fact_2(base)\n")
    test.startswith(stderr.getvalue(), "")


@test
def clingo_dropin_default_hook_errors(tmp_path, argv, stdout, stderr):
    f = tmp_path/'f'
    f.write_text("""syntax error """)
    argv += [f.as_posix(), '--run-python-tests']
    with test.raises(SyntaxError, "syntax error, unexpected <IDENTIFIER>") as e:
        clingo_plus()
    test.contains(stdout.getvalue(), """UNKNOWN\n
Models       : 0+""")
    test.eq(
        "    1 syntax error \n             ^^^^^ syntax error, unexpected <IDENTIFIER>",
        e.exception.text)
    test.eq(stderr.getvalue(), '')


@test
def access_python_script_functions(tmp_path, argv, stdout, stderr):
    f = tmp_path/'f'
    f.write_text("""
#script (python)
def my_func(a):
    return a
#end.
#program test_one.
something(@my_func("hello")).
models(1).
    """)
    argv += [f.as_posix(), '--run-asp-tests', '--run-python-tests']
    clingo_plus()
    s = stdout.getvalue()
    test.contains(s, f"Testing {f}\n  base()\n  test_one()\n")
    test.startswith(stderr.getvalue(), "")


@test
def bug_read_stdin_and_solve_with_run_tests():
    # solving was prevented because stdin was read by the tester and gone for the next step
    p = spawn_clingo_plus(input=b"a. b. c.", arguments=['--run-python-tests'])
    test.contains(p.stdout, b"Answer: 1 (Time: 0.001s)\na b c\nSATISFIABLE\n")
    p = spawn_clingo_plus(input=b"a. b. c.", arguments=[])
    test.contains(p.stdout, b"Answer: 1 (Time: 0.001s)\na b c\nSATISFIABLE\n")
    p = spawn_clingo_plus(input=b"a. b. c.", arguments=['--run-python-tests', '--run-asp-tests'])
    test.contains(p.stdout, b"Answer: 1 (Time: 0.001s)\na b c\nSATISFIABLE\n")
    p = spawn_clingo_plus(input=b"a. b. c.", arguments=['--run-asp-tests'])
    test.contains(p.stdout, b"Answer: 1 (Time: 0.001s)\na b c\nSATISFIABLE\n")
    test.eq(b'', p.stderr)



import shutil
import enum
import itertools
import tempfile
from unittest import mock
import clingo

import selftest
test = selftest.get_tester(__name__)


NA = clingo.String("N/A")

            
# ExitCode, see clasp_app.h
class ExitCode(enum.IntEnum):
    UNKNOWN   =   0  #/*!< Satisfiability of problem not known; search not started.   */
    INTERRUPT =   1  #/*!< Run was interrupted.                                       */
    SAT       =  10  #/*!< At least one model was found.                              */
    EXHAUST   =  20  #/*!< Search-space was completely examined.                      */
    MEMORY    =  33  #/*!< Run was interrupted by out of memory exception.            */
    ERROR     =  65  #/*!< Run was interrupted by internal error.                     */
    NO_RUN    = 128  #/*!< Search not started because of syntax or command line error.*/
 

def write_file(file, text):
    file.write_text(text)
    return file.as_posix()


def write_tempfile(suffix, data):
    f = tempfile.NamedTemporaryFile('w', suffix=suffix)
    f.write(data)
    f.flush()
    return f


def list_symbols(control):
    return [str(a.symbol) for a in control.symbolic_atoms]


def create_control(arguments=(), logger=None, **etc):
    new_args=list(itertools.dropwhile(lambda p: not p.startswith('--'), arguments))
    return clingo.Control(arguments=new_args, logger=logger)


CR = '\n' # trick to support old python versions that do not accecpt \ in f-strings

def batched(iterable, n):
    """ not in python < 3.12 """
    # batched('ABCDEFG', 3) → ABC DEF G
    if n < 1:
        raise ValueError('n must be at least one')
    iterator = iter(iterable)
    while batch := tuple(itertools.islice(iterator, n)):
        yield batch


@test
def batch_it():
    test.eq([], list(batched([], 1)))
    test.eq([(1,)], list(batched([1], 1)))
    test.eq([(1,),(2,)], list(batched([1,2], 1)))
    test.eq([(1,)], list(batched([1], 2)))
    test.eq([(1,2)], list(batched([1,2], 2)))
    test.eq([(1,2), (3,)], list(batched([1,2,3], 2)))
    with test.raises(ValueError, 'n must be at least one'):
        list(batched([], 0))


def format_symbols(symbols):
    symbols = sorted(str(s).strip()for s in symbols)
    if len(symbols) > 0:
        col_width = (max(len(w) for w in symbols)) + 2
        width, h = shutil.get_terminal_size((120, 20))
        cols = width // col_width
        modelstr = '\n'.join(
                ''.join(s.ljust(col_width) for s in b).strip()
            for b in batched(symbols, max(cols, 1)))
    else:
        modelstr = "<empty>"
    return modelstr


@test
def format_symbols_basic():
    test.eq('a', format_symbols(['a']))
    test.eq('a  b  c  d', format_symbols(['a', 'b', 'c', 'd']))
    test.eq('a  b  c  d', format_symbols([' a  ', '\tb', '\nc\n', '  d '])) # strip
    with mock.patch("shutil.get_terminal_size", lambda _: (10,20)):
        test.eq('a  b  c\nd', format_symbols(['a', 'b', 'c', 'd']))
    with mock.patch("shutil.get_terminal_size", lambda _: (8,20)):
        test.eq('a  b\nc  d', format_symbols(['a', 'b', 'c', 'd']))
    with mock.patch("shutil.get_terminal_size", lambda _: (4,20)):
        test.eq('a\nb\nc\nd', format_symbols(['a', 'b', 'c', 'd']))


def is_plugin_instruction(p, givenname='insert_plugin'):
    """ check if p is a processor(<classname>) and return <classname> """
    if p.ast_type == clingo.ast.ASTType.Rule:
        p = p.head
        if p.ast_type == clingo.ast.ASTType.Literal:
            p = p.atom
            if p.ast_type == clingo.ast.ASTType.SymbolicAtom:
                p = p.symbol
                if p.ast_type == clingo.ast.ASTType.Function:
                    name, args = p.name, p.arguments
                    if name == givenname:
                        p = args[0]
                        if p.ast_type == clingo.ast.ASTType.SymbolicTerm:
                            p = p.symbol
                            if p.type == clingo.symbol.SymbolType.String:
                                p = p.string
                                if isinstance(p, str):
                                    return p





import os
import itertools
import clingo

from .messageparser import warn2raise
from .misc import write_file

import selftest
test = selftest.get_tester(__name__)


def msg2exc(code, message):
    return warn2raise(None, None, code, message)


def clingo_syntaxerror_plugin(next, msg2exc=msg2exc, **etc):
    """ Takes clingo log message to raise rich exception."""

    _logger, _main = next(**etc)

    exceptions = []

    def logger(code, message):
        #_logger(code, message)
        rich_exception = msg2exc(code, message)
        exceptions.append(rich_exception)

    def main():
        try:
            result = _main() # expect Clingo to call logger on error
        except RuntimeError as e:
            if not exceptions:
                raise e
        if exceptions:
            notes = ((s, len(tuple(n))) for s, n in itertools.groupby(exceptions[1:], lambda e: f"followed by: {e}"))
            x = [f"{s} (repeated {n} times)" if n > 1 else s for s, n in notes]
            exceptions[0].__notes__ = x
            raise exceptions[0]
        return result
            
    return logger, main


@test
def syntaxerror_basics():
    
    def next(control=None):
        def next_main():
            control(42, "not good") # fake; but we only want to trigger logger
            raise RuntimeError
        def next_logger(code, message):
            pass
        return next_logger, next_main
        
    # we fake the control and directly pass the logger
    def prime_logger(code, message):
        logger(code, message)
            
    logger, main = clingo_syntaxerror_plugin(next, control=prime_logger, msg2exc=lambda code, msg: SyntaxError(msg))
        
    with test.raises(SyntaxError, "not good"):
        main()

    # TODO somehow test mapping the code/message to a fully specified
    #      SyntaxError separately (mapping code in messageparser.py)
    #      - test passing etc
    #      - test calling next_logger
    # warn2raise(lines, label, code, msg)


def run_syntaxerror_plugin(tmp_path, source=None, label='test', files=(), parts=(('base', ()),)):
    if source:
        f = write_file(tmp_path/label, source)
        files = (f, *files)
    def next(**_):
        def load_and_ground():
            for f in files:
                control.load(f)
            control.ground()
            return "done"
        return lambda x,y: None, load_and_ground
    logger, main = clingo_syntaxerror_plugin(next)
    control = clingo.Control(logger=logger)
    return main()


@test
def return_result_when_no_error(tmp_path):
    result = run_syntaxerror_plugin(tmp_path, "a.")
    test.eq('done', result)


@test
def exc_with_label(tmp_path):
    with test.raises(SyntaxError, "syntax error, unexpected <IDENTIFIER>") as e:
        run_syntaxerror_plugin(tmp_path, "a.\nan error", label='my code')
    test.eq("""    1 a.
    2 an error
         ^^^^^ syntax error, unexpected <IDENTIFIER>""", e.exception.text)
    test.endswith(e.exception.filename, 'my code')


@test
def exception_in_included_file(tmp_path):
    f = tmp_path/'error.lp'
    f.write_text("error")
    old = os.environ.get('CLINGOPATH')
    try:
        os.environ['CLINGOPATH'] = tmp_path.as_posix()
        with test.raises(SyntaxError, 'syntax error, unexpected EOF') as e:
            run_syntaxerror_plugin(tmp_path, """#include "error.lp".""", label='main.lp')
        test.eq(f.as_posix(), e.exception.filename)
        test.eq(2, e.exception.lineno)
        test.eq('    1 error\n      ^ syntax error, unexpected EOF', e.exception.text)
    finally:
        if old:  #pragma no cover
            os.environ['CLINGOPATH'] = old


@test
def parse_warning_raise_error(tmp_path):
    with test.raises(SyntaxError, "syntax error, unexpected EOF") as e:
        run_syntaxerror_plugin(tmp_path, "abc", label='code_a')
    test.endswith(e.exception.filename, 'code_a')
    test.eq(2, e.exception.lineno)
    test.eq("    1 abc\n      ^ syntax error, unexpected EOF", e.exception.text)

    with test.raises(SyntaxError, 'atom does not occur in any rule head:  b') as e:
        run_syntaxerror_plugin(tmp_path, "a :- b.")
    test.endswith(e.exception.filename, 'test')
    test.eq(1, e.exception.lineno)
    test.eq("    1 a :- b.\n           ^ atom does not occur in any rule head:  b", e.exception.text)

    with test.raises(SyntaxError, 'operation undefined:  ("a"/2)') as e:
        run_syntaxerror_plugin(tmp_path, 'a("a"/2).')
    test.endswith(e.exception.filename, 'test')
    test.eq(1, e.exception.lineno)
    test.eq('    1 a("a"/2).\n        ^^^^^ operation undefined:  ("a"/2)',
            e.exception.text)

    with test.raises(SyntaxError, "unsafe variables in:  a(A):-[#inc_base];b.") as e:
        run_syntaxerror_plugin(tmp_path, 'a(A)  :-  b.', label='code b')
    test.endswith(e.exception.filename, 'code b')
    test.eq(1, e.exception.lineno)
    test.eq("""    1 a(A)  :-  b.
        ^ 'A' is unsafe
      ^^^^^^^^^^^^ unsafe variables in:  a(A):-[#inc_base];b.""",
            e.exception.text)

    with test.stdout as out:
        with test.raises(SyntaxError, "global variable in tuple of aggregate element:  X") as e:
            run_syntaxerror_plugin(tmp_path, 'a(1). sum(X) :- X = #sum { X : a(A) }.')
        test.endswith(e.exception.filename, 'test')
        test.eq(1, e.exception.lineno)
        test.eq("""    1 a(1). sum(X) :- X = #sum { X : a(A) }.
                                 ^ global variable in tuple of aggregate element:  X""",
                e.exception.text)
        test.eq(['followed by: unsafe variables in:  sum(X):-[#inc_base];X=#sum{X:a(A)}. (test, line 1)'],
            e.exception.__notes__)
        #test.startswith(
        #        out.getvalue(),
        #        "  WARNING ALREADY exception: global variable in tuple of aggregate element:  X (test, line 1)")


@test
def unsafe_variables(tmp_path):
    with test.raises(SyntaxError, "unsafe variables in:  output(A,B):-[#inc_base];input.") as e:
        run_syntaxerror_plugin(tmp_path, """
            input.
            output(A, B)  :-  input.
            % comment""")
    test.endswith(e.exception.filename, 'test')
    test.eq(3, e.exception.lineno)
    test.eq("""    1 
    2             input.
    3             output(A, B)  :-  input.
                         ^ 'A' is unsafe
                            ^ 'B' is unsafe
                  ^^^^^^^^^^^^^^^^^^^^^^^^ unsafe variables in:  output(A,B):-[#inc_base];input.
    4             % comment""", e.exception.text)


@test
def multiline_error(tmp_path):
    with test.raises(SyntaxError,
                     "unsafe variables in:  geel(R):-[#inc_base];iets_vrij(S);(S,T,N)=R;R=(S,T,N)."
                     ) as e:
        run_syntaxerror_plugin(tmp_path, """
            geel(R)  :-
                iets_vrij(S), R=(S, T, N).
            %%%%""")
    test.endswith(e.exception.filename, 'test')
    test.eq(3, e.exception.lineno)
    test.eq("""    1 
    2             geel(R)  :-
                       ^ 'R' is unsafe
    3                 iets_vrij(S), R=(S, T, N).
                                          ^ 'T' is unsafe
                                             ^ 'N' is unsafe
                  ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ unsafe variables in:  geel(R):-[#inc_base];iets_vrij(S);(S,T,N)=R;R=(S,T,N).
    4             %%%%""", e.exception.text)


@test
def duplicate_const(tmp_path):
    with test.raises(SyntaxError, "redefinition of constant:  #const a=43.") as e:
        run_syntaxerror_plugin(tmp_path, """
            #const a = 42.
            #const a = 43.
            """, parts=[('base', ()), ('p1', ()), ('p2', ())])
    test.endswith(e.exception.filename, 'test')
    test.eq(3, e.exception.lineno)
    test.eq("""    1 
    2             #const a = 42.
                  ^^^^^^^^^^^^^^ constant also defined here
    3             #const a = 43.
                  ^^^^^^^^^^^^^^ redefinition of constant:  #const a=43.
    4             """, e.exception.text, diff=test.diff)


@test
def error_in_file(tmp_path):
    code = tmp_path/'code.lp'
    code.write_text('oops(().')
    with test.raises(SyntaxError) as e:
        with run_syntaxerror_plugin(tmp_path, files=[code.as_posix()]) as s:
            s.go_prepare()
        test.endswith(e.exception.text, """
    1 oops(().
             ^ syntax error, unexpected ., expecting ) or ;""")


@test
def do_not_mask_other_exceptions(stdout):

    def next_plugin():
        def main():
            raise Exception("do not mask me")
        return None, main
       
    _, main = clingo_syntaxerror_plugin(next_plugin)
    with test.raises(Exception, "do not mask me"):
        main()


@test
def detect_runtimerror_without_logged_error(stdout):

    def next_plugin():
        def main():
            raise RuntimeError("no logger called")
        return None, main

    _, main = clingo_syntaxerror_plugin(next_plugin)
    with test.raises(RuntimeError, "no logger called"):
        main()


@test
def raise_exception_in_context():

    class Context:
        def f(self):
            1/0

    def next_plugin():
        def main():
            control.add("a(@f()).")
            control.ground(context=Context())
        return lambda c, m: None, main

    logger, main = clingo_syntaxerror_plugin(next_plugin)
    control = clingo.Control(logger=logger)
    with test.raises(ZeroDivisionError, "division by zero"):
        main()


@test
def count_repeated_errors():
    def next_plugin():
        def main():
            control.add("a(@f()). b(@f()). c(@f()).")
            control.ground()
        return lambda c, m: None, main
    logger, main = clingo_syntaxerror_plugin(next_plugin)
    control = clingo.Control(logger=logger)
    with test.raises(SyntaxError, "operation undefined:  function 'f' not found") as e:
        main()
    test.eq(["followed by: operation undefined:  function 'f' not found (<asp code>, line 1) (repeated 2 times)"], e.__notes__)




import traceback
import re
import os
import math
import pathlib

import clingo

import selftest
test = selftest.get_tester(__name__)


CR = '\n' # trick to support old python versions that do not accecpt \ in f-strings


class AspSyntaxError(SyntaxError):
    pass


def parse_message(msg):
    return [(file,
             int(end_line_or_col if opt_end_col else start_line),
             int(start_col),
             int(opt_end_col if opt_end_col else end_line_or_col),
             key, msg, more) 
            for file, start_line, start_col, end_line_or_col, opt_end_col, key, msg, more
            in re.findall(r"(?m)^(.+?):(\d+):(\d+)-(\d+)(?::(\d+))?:\s(.+?):\s([^\n]+)(?:\n(\s\s.+))?", msg)]


@test
def parse_clingo_error_messages():
    test.eq([('<block>', 1, 6, 7, 'info', 'atom does not occur in any rule head:', '  b')],
            parse_message("<block>:1:6-7: info: atom does not occur in any rule head:\n  b"))
    test.eq([('<block>', 1, 4, 9, 'error', 'syntax error, unexpected <IDENTIFIER>', '')],
            parse_message("<block>:1:4-9: error: syntax error, unexpected <IDENTIFIER>"))
    test.eq([('/var/folders/fn/2hl6h1jn4772vw7j9hlg9zjm0000gn/T/tmpfy706dra/error.lp', 2, 1, 2,
              'error', 'syntax error, unexpected EOF', '')],
            parse_message("/var/folders/fn/2hl6h1jn4772vw7j9hlg9zjm0000gn/T/tmpfy706dra/error.lp:2:1-2:"
                          " error: syntax error, unexpected EOF"))
    test.eq([('<block>', 1, 3, 8, 'info', 'operation undefined:', '  ("a"/2)')],
            parse_message('<block>:1:3-8: info: operation undefined:\n  ("a"/2)'))
    test.eq([('<blOck>', 1, 1, 11, 'error', 'unsafe variables in:', '  a(A):-[#inc_base];b.'),
             ('<block>', 1, 3, 4, 'note', "'A' is unsafe", '')],
            parse_message("""<blOck>:1:1-11: error: unsafe variables in:
  a(A):-[#inc_base];b.
<block>:1:3-4: note: 'A' is unsafe"""))
    test.eq([('<block>', 1, 7, 39, 'error', 'unsafe variables in:', '  sum(X):-[#inc_base];X=#sum{X:a(A)}.'),
             ('<block>', 1, 11, 12, 'note', "'X' is unsafe", '')],
            parse_message("""<block>:1:7-39: error: unsafe variables in:
  sum(X):-[#inc_base];X=#sum{X:a(A)}.
<block>:1:11-12: note: 'X' is unsafe"""))
    test.eq([('<block>', 3, 13, 37, 'error', 'unsafe variables in:', '  output(A,B):-[#inc_base];input.'),
             ('<block>', 3, 20, 21, 'note', "'A' is unsafe", ''),
             ('<block>', 3, 23, 24, 'note', "'B' is unsafe", '')],
            parse_message("""<block>:3:13-37: error: unsafe variables in:
  output(A,B):-[#inc_base];input.
<block>:3:20-21: note: 'A' is unsafe
<block>:3:23-24: note: 'B' is unsafe"""))
    test.eq([('<block>', 3, 13, 43, 'error', 'unsafe variables in:', '  geel(R):-[#inc_base];iets_vrij(S);(S,T,N)=R;R=(S,T,N).'),
             ('<block>', 3, 40, 41, 'note', "'N' is unsafe", ''),
             ('<block>', 2, 18, 19, 'note', "'R' is unsafe", ''),
             ('<block>', 3, 37, 38, 'note', "'T' is unsafe", '')
             ],
            parse_message("""<block>:2:13-3:43: error: unsafe variables in:
  geel(R):-[#inc_base];iets_vrij(S);(S,T,N)=R;R=(S,T,N).
<block>:3:40-41: note: 'N' is unsafe
<block>:2:18-19: note: 'R' is unsafe
<block>:3:37-38: note: 'T' is unsafe"""), diff=test.diff)



def warn2raise(lines, label, code, msg):
    """ Clingo calls this, but can't handle exceptions well, so we wrap everything. """
    try:
        # deal with '<cmd>' (command line) error messages separately
        if msg and msg.startswith('<cmd>'):
            return RuntimeError(msg)
        messages = parse_message(msg)
        file, line, start, end, key, msg, more = messages[0]
        name = label if label else '<asp code>'
        srclines = lines if lines else []
        if file == '<block>':
            srclines = lines if lines else []
        elif pathlib.Path(file).exists():
            name = file
            srclines = [l.removesuffix('\n') for l in open(file).readlines()]
        w = 1
        max_lineno = len(srclines)
        nr_width = 1 + int(math.log10(max_lineno)) if max_lineno > 0 else 0
        srclines = [f"    {n:{nr_width}} {line}" for n, line in enumerate(srclines, 1)]
        msg_fmt = lambda: f"   {' ':{nr_width}}  {' ' * (start-1)}{'^' * (end-start)} {m}{r}"
        offset = 0
        for _, line, start, end, _, m, r in sorted(messages[1:]):
            srclines.insert(line + offset, msg_fmt())
            offset += 1 
        _, line, start, end, _, m, r = messages[0]
        srclines.insert(line + len(messages) -1, msg_fmt())
        snippet = srclines[max(0,line-10):line+10]  # TODO testme

        if "file could not be opened" in m:
            snippet.append(f"CLINGOPATH={os.environ.get('CLINGOPATH')}")
        se = AspSyntaxError(msg + r, (name, line, None, CR.join(snippet)))
        return se
    except BaseException as e:
        """ unexpected exception in the code above """
        traceback.print_exc()
        exit(-1)


# NB: most tests dor warn2raise are still in runasptests.py, indirectly checking for SyntaxError's
@test
def raise_warnings_as_exceptions(stderr):
    try:
        warn2raise(None, None, None, None)
    except SystemExit as e:
        test.eq(-1, e.code)
    msg = stderr.getvalue()
    test.startswith(msg, "Traceback (most recent call last):")
    test.endswith(msg, "TypeError: expected string or bytes-like object, got 'NoneType'\n")


@test
def deal_with_command_line_errors():
    e = warn2raise(None, None, None, "<cmd>: all wrong!")
    test.eq(RuntimeError, type(e))
    test.eq(('<cmd>: all wrong!',), e.args)


@test
def print_clingo_path_on_file_could_not_be_opened():
    old = os.environ.get('CLINGOPATH', None)
    os.environ['CLINGOPATH'] = 'paf'
    try:
        error = warn2raise([], 'no-file', None, "<block>:3:4-5: info: file could not be opened:\n  dus")
        test.isinstance(error, AspSyntaxError)
        test.eq("no-file", error.filename)
        test.eq(3, error.lineno)
        test.eq(None, error.offset)
        test.eq('         ^ file could not be opened:  dus\nCLINGOPATH=paf', error.text)
        test.eq('file could not be opened:  dus', error.msg)
    finally:
        if old:
            os.environ['CLINGOPATH'] = old


@test
def make_snippet_witg_proper_indent():
    src_lines = ['%1', '%2','%3','%4','%5','%6','%7','%8','%9','%10','error']
    error = warn2raise(src_lines, None, 'code', "<block>:11:1-6: info: hier moet een punt staan:\n  snappie?")
    test.eq("""     2 %2
     3 %3
     4 %4
     5 %5
     6 %6
     7 %7
     8 %8
     9 %9
    10 %10
    11 error
       ^^^^^ hier moet een punt staan:  snappie?""",
        error.text, diff=test.diff)

@test
def nieuw_geval():
    source = """a.
an error"""
    error = warn2raise(source.splitlines(), None, 'code', '<string>:2:4-9: error: syntax error, unexpected <IDENTIFIER>')
    test.eq("""    1 a.
    2 an error
         ^^^^^ syntax error, unexpected <IDENTIFIER>""", error.text)







from .misc import write_file

import selftest

test = selftest.get_tester(__name__)


def clingo_sequencer_plugin(
        next,
        control=None,
        files=(),
        parts=(('base', ()),),
        context=None,
        on_model=None,
        yield_=False,
        **etc):
    """ Breaks down main into Clingo-specific steps. """
    
    logger, load, ground, solve = next(parts=parts, **etc)
            
    def main():
        load(control, files=files)
        ground(control, parts=parts, context=context)
        return solve(control, on_model=on_model, yield_=yield_)
            
    return logger, main



@test
def sequencer_plugin_basics(tmp_path):

    logs = []
    trace = []

    def next_plugin(logger=None, **etc):
        trace.append(etc)
        def load(control, files=()):
            trace.append(files)
        def ground(control, parts=(('base', ()),), context=None):
            trace.append(parts)
            trace.append(context)
        def solve(control, on_model=None, yield_=False):
            # many more, see https://potassco.org/clingo/python-api/current/clingo/control.html#clingo.control.Control.solve
            trace.append(on_model)
            trace.append(yield_)
        return logger, load, ground, solve

    def my_logger(code, message):
        logs.append((code, message))

    class MyContext:
        def my_func(a):
            return a

    def my_on_model(model):
        pass

    file1 = write_file(tmp_path/'file1.lp', "a(@my_func(b)).")
    logger, main = clingo_sequencer_plugin(
        next_plugin,
        logger=my_logger,
        files=(file1,),
        parts=(('part_a', ()), ('part_b', ())),
        context=MyContext(),
        on_model=my_on_model,
        yield_=True,
        more='better')

    logger(42, "message 42")
    test.eq([(42, "message 42")], logs)

    main()

    test.eq({'parts': (('part_a', ()), ('part_b', ())), 'more': 'better'}, trace[0])
    test.eq((file1,), trace[1])
    test.eq((('part_a', ()), ('part_b', ())), trace[2])
    test.isinstance(trace[3], MyContext)
    test.eq(my_on_model, trace[4])
    test.eq(True, trace[5])


@test
def test_defaults():
    
    trace = []
        
    def next_plugin(logger=None, **etc):
        trace.append(etc)
        def load(control, files='-'):
            trace.append(files)
        def ground(control, parts='-', context='-'):
            trace.append(parts)
            trace.append(context)
        def solve(control, on_model='-', yield_='-'):
            trace.append(on_model)
            trace.append(yield_)
        return logger, load, ground, solve

    #file1 = write_file(tmp_path/'file1.lp', "a(@my_func(b)).")
    logger, main = clingo_sequencer_plugin(next_plugin)

    main()

    test.eq({'parts': (('base', ()),)}, trace[0])
    test.eq((), trace[1])
    test.eq((('base', ()),), trace[2])
    test.eq(None, trace[3])
    test.eq(None, trace[4])
    test.eq(False, trace[5])




""" Separate module to allow inspecting args before running selftests """

import argparse
import sys
import selftest


silent = argparse.ArgumentParser(add_help=False, exit_on_error=False)
silent.add_argument('--run-python-tests', help="Also run in-source Python tests.", action='store_true')


def maybe_silence_tester(argv=None):
    args, unknown = silent.parse_known_args(argv)
    if not args.run_python_tests:
        try:
            # must be called first and can only be called once, but, when
            # we are imported from another app that also uses --silent, 
            # that app might already have called basic_config()
            # TODO testme
            selftest.basic_config(run=False)
        except AssertionError:
            root = selftest.get_tester(None)
            CR = '\n'
            assert not root.option_get('run'), "In order to NOT run Python tests, " \
                f"Tester {root}{CR} must have been configured to NOT run tests."
    return unknown


def parse_plus_arguments(argv=None):
    argparser = argparse.ArgumentParser(
            parents=[silent],
            add_help=False,
            exit_on_error=False,
            description='Runs in-source ASP tests in given logic programs, on top of standard clingo.',
            epilog="Clingo options below.\n")
    argparser.add_argument('--run-asp-tests', help="Run all selftests in ASP code.", action='store_true')
    # we try to make the --help as compatible with Clingo as possible
    argparser.add_argument('-h', '--help', help="Show all info on arguments.", type=int, nargs='?', choices=(1,2,3), const=1, default=None)
    args, unknown = argparser.parse_known_args(argv)
    if args.help:
        argparser.print_help()
        unknown.insert(0, f'--help={args.help}') # pass it to Clingo as well
    return args, unknown




import os
import subprocess
import tempfile
import pathlib

from .misc import write_tempfile

import selftest
test = selftest.get_tester(__name__)


def stdin_to_tempfile_plugin(next, files=(), **etc):
    """ Writes stdin to a temporary file so it can be read by multiple plugins """

    stdinput_file = None

    if not files:
        stdinput = open(os.dup(0)).read()
        stdinput_file = write_tempfile('-stdin.lp', stdinput)
        files = [stdinput_file.name]

    logger, _main = next(files=files, **etc)

    def main():
        stdinput_file   # keep save from GC
        return _main()

    return logger, main


def run_test():
    # NB: runs in another process!
    def next_plugin(files=(), **etc):
        def main():
            data = open(files[0]).read()
            return files, data
        return 42, main
    logger, main = stdin_to_tempfile_plugin(next_plugin, files=())
    test.eq(42, logger)
    files, data = main()
    for f in files:
        print(f)
    print(data)


@test
def turn_stdin_into_tempfile():
    # test in another process since we cannot mock stdin when read from Clingo's C++ runtime
    path = pathlib.Path(__file__).parent
    p = subprocess.run(
        ["python", "-c", f"from asp_selftest.plugins.stdin_to_tempfile_plugin import run_test; run_test()"],
        env=os.environ | {'PYTHONPATH': path},
        input=b"asp. is. nice.",
        capture_output=True)
    test.eq(b'', p.stderr)
    output_lines = p.stdout.splitlines()
    test.startswith(output_lines[0], tempfile.gettempdir().encode())
    test.endswith(output_lines[0], b"-stdin.lp")
    test.eq(b"asp. is. nice.", output_lines[1])
    test.eq(2, len(output_lines))




import sys
import clingo
import pathlib

from .asputil import is_tuple, is_function, mk_symbol, mk_theory_atom
from ..misc import write_file, create_control, list_symbols

import selftest
test = selftest.get_tester(__name__)


""" Support for reification of rules in ASP code.
    It supports predicates and a special theory:

      rule(<head>, <body0>, ... <bodyN>).
      &rule(<head>) { body0; ...; <bodyN>).

    The predicate form can be handy when the program must remain ASP-compatible.
    Be careful: it can be used in both head and body, while only head makes sense.
    The theory form is restricted to heads and can handle conditional bodies (variable bodies).
    The latter is the prefered method.

    A tuple (<name>, <arg0>, ..., <argN>) if will transform to:

      <name>(<arg0>, ..., <argN>).

    When <name> is itself a function, its arguments are prepended to <arg0>..<argN>.
    Thus, (name(a), b, c) becomes name(a, b, c).
"""

# ASP #theory defining the symtax of the &reify/1 atom
MY_PATH = pathlib.Path(__file__).resolve()
THEORY_PATH = MY_PATH.parent
THEORY_FILE = MY_PATH.with_name('reify.lp')


def clingo_reify_plugin(
        next,
        on_rule=lambda _:None,
        parts=(('base', ()),),
        context=None,
        arguments=(),
        **etc):
    """ Plugin turning rule predicates into reallified rules and adds them to the control."""

    logger, _load, ground, solve = next(parts=parts, context=context, arguments=arguments, **etc)  # test **etc

    def load(control, files):
        reground = True
        rules_added = set()

        while reground:
            reground = False
            sub_control = create_control(arguments=[*arguments, '--warn', 'no-atom-undefined'], **etc)
            _load(sub_control, files)
            
            for rule in rules_added:
                sub_control.add(rule)
            ground(sub_control, parts=parts, context=context)
                        
            for rule in reified_rules(sub_control):
                if rule not in rules_added:
                    rules_added.add(rule)
                    sub_control.add(rule)
                    on_rule(rule)
                    reground = True
                                
        _load(control, files)
                    
        for rule in rules_added:
            control.add(rule)

    return logger, load, ground, solve


def to_symbol(theory_term):
    if isinstance(theory_term, clingo.TheoryTerm):
        return clingo.parse_term(str(theory_term))
    return theory_term


def make_function(arguments):
    name, *args = arguments
    #assert is_string(name) or is_function(name), \
    #       f"First element '{name}' must be a name or function {name.type}"
    try:
        args = args + name.arguments # + args
    except RuntimeError:
        pass
    try:
        name = name.name
    except RuntimeError:
        name = str(name)[1:-1]
    args = [to_symbol(a) for a in args]
    return clingo.Function(name, args)


@test
def make_function_from_symbols():
    T = "#theory x{t{};&sym/1:t,head;&sym/2:t,head}."
    test.eq('naam', str(make_function([mk_symbol('naam')])))
    test.eq('naam(a)', str(make_function([mk_symbol('naam'), mk_symbol('a')])))
    test.eq('naam(b,a)', str(make_function([mk_symbol('naam(a)'), mk_symbol('b')])))

    with mk_theory_atom('&sym(naam).', T) as a:
        test.eq('naam', str(make_function(a.term.arguments)))
    with mk_theory_atom('&sym(naam(a)).', T) as a:
        test.eq('naam(a)', str(make_function(a.term.arguments)))
    with mk_theory_atom('&sym(naam(a), b).', T) as a:
        test.eq('naam(b,a)', str(make_function(a.term.arguments)))


def reified_rules(control):
    """ Reads rule predicates from the control and returns reified ASP rules. """

    def reifies():
        by_signature = control.symbolic_atoms.by_signature
        for sa in by_signature('rule', 1):
            yield sa.symbol, sa.symbol.arguments
        for sa in by_signature('rule', 2):
            yield sa.symbol, sa.symbol.arguments
        for ta in control.theory_atoms:
            term = ta.term
            if term.name.startswith('rule'):
                yield ta, term.arguments + ta.elements

    for function, arguments in reifies():
        assert arguments, f"'{function}' must at least have one argument"
        head, *body = arguments
        if is_tuple(head):
            head = make_function(head.arguments)

        if body:
            if isinstance(body[0], clingo.TheoryElement):
                new_body = []
                for b in body:
                    for t in b.terms:
                        if is_tuple(t):
                            f = make_function(t.arguments)
                            new_body.append(f)
                        else:
                            new_body.append(t)
                body = new_body
            yield f"{head} :- {', '.join(map(str, body))}.\n"
        else:
            yield f"#external {head}.\n"


# BELOW SOME TESTS FOR INSTANTIATING RULES

def test_reified_rules(asp, reified, **etc):
    """ Helper for testing reified_rules. """
    if "&rule" in asp:
        asp += f'#include "{THEORY_FILE}".\n'
    control = clingo.Control(arguments=['--warn', 'no-atom-undefined'])
    control.add(asp)
    control.ground(**etc)
    new_rules = reified_rules(control)
    test.eq(reified.strip(), ''.join(new_rules).strip())
    return control


@test
def simple_fact_predicate():
    test_reified_rules(
"""
def(f(42)).
rule(A) :- def(A).
""", """
#external f(42).
""")


@test
def simple_fact_theory():
    test_reified_rules(
"""
def(f(42)).
&rule(A) :- def(A).
""", """
#external f(42).
""")


@test
def simple_rule(stderr):
    test_reified_rules(
"""
rule(f(41), g(42)).
b.
g(42) :- b.
c :- f(41).
""", """
f(41) :- g(42).
""")
    test.eq('', stderr.getvalue())


@test
def rule_with_condition():
    test_reified_rules(
"""
&rule(head(41)) { body0(42);  body1(N) : N=43..44 }.
""", """
head(41) :- body0(42), body1(43), body1(44).
""")


@test
def head_variable_theory():
    test_reified_rules(
"""
step(stuur(links)).
&rule(A) { body0(42) }  :-  step(A).
""", """
stuur(links) :- body0(42).
""")


@test
def head_variable_predicate():
    test_reified_rules(
"""
step(stuur(links)).
rule(A, body0(42))  :-  step(A).
""", """
stuur(links) :- body0(42).
""")


@test
def head_function():
    test_reified_rules(
"""
define(stuur).
rule((F), body0(42))  :-  define(F).
""", """
stuur :- body0(42).
""")


@test
def head_function_with_arg():
    test_reified_rules(
"""
define(stuur(links)).
rule((F), body0(42))  :-  define(F).
""", """
stuur(links) :- body0(42).
""")


@test
def head_function_with_additional_args():
    test_reified_rules(
"""
define(stuur(links, now)).
rule((F, 2, 3), body0(42))  :-  define(F).
""", """
stuur(2,3,links,now) :- body0(42).
""")


@test
def head_function_with_string():
    test_reified_rules(
"""
rule(("aap", 2, 3), body0(42)).
""", """
aap(2,3) :- body0(42).
""")


@test
def head_function_with_symbol():
    test_reified_rules(
"""
rule((aap, 2, 3), body0(42)).
""", """
aap(2,3) :- body0(42).
""")


@test
def tuples_in_theory():
    test_reified_rules(
"""
&rule((a, 1, 2)) { (b, 3),  (c, 4),  d(5) }.
""", """
a(1,2) :- b(3), c(4), d(5).
""")


@test
def reify_with_context(stderr):
    class Context:
        @staticmethod
        def zeep():
            return clingo.String("sop")
    control = test_reified_rules("""
                         b(@zeep).
                         a.
                         rule(geel, a).
                         """,
                         "geel :- a.",
                         context=Context())
    symbols = {str(sa.symbol) for sa in control.symbolic_atoms}
    test.contains(symbols, 'b("sop")')


# BELOW SOME TESTS FOR THE PLUGIN

def test_reify_plugin(tmp_path, code, rules, trace=lambda _:None, arguments=(), **etc):
    """ Helper for testing the plugin. """
    f = write_file(tmp_path/'f.lp', code)
    def next_plugin(**etc):
        def load(control, files):
            trace(('load', control, files))
            for f in files:
                control.load(f)
        def ground(control, **etc):
            trace(('ground', control, etc))
            control.ground(**etc)
        return None, load, ground, None
    new_rules = set()
    _, load, ground, _ = clingo_reify_plugin(next_plugin, on_rule=new_rules.add, arguments=arguments, **etc)
    control = clingo.Control(arguments=arguments)
    load(control, (f,))
    test.eq(rules, new_rules)
    ground(control, **etc)
    return control, ground


@test
def reify_plugin_basics(tmp_path):
    trace = []
    control, ground = test_reify_plugin(tmp_path, "b. rule(a, b).", {"a :- b.\n"}, trace.append)

    trace = iter(trace) 
    l, c0, f = next(trace)  # test if it uses new control with next_plugin for load
    test.eq('load', l)
    test.ne(control, c0)
    test.eq(1, len(f))
    test.endswith(f[0], '/f.lp')

    g, c1, etc = next(trace)  # test if it uses new control with next_plugin for ground
    test.eq('ground', g)
    test.ne(control, c1)
    test.eq(c0, c1)
    test.eq((('base', ()),), etc['parts'])
    test.eq(None, etc['context'])

    l, c0, f = next(trace)  # second load
    test.eq('load', l)
    test.ne(control, c0)
    test.eq(1, len(f))
    test.endswith(f[0], '/f.lp')

    g, c2, etc = next(trace)  # second ground for ensuring end of propagation
    test.eq('ground', g)
    test.ne(control, c2)
    test.eq(c0, c2)
    test.eq((('base', ()),), etc['parts'])
    test.eq(None, etc['context'])

    l, c2, f = next(trace)  # test if is finally load the code into our control
    test.eq('load', l)
    test.eq(control, c2)
    test.eq(1, len(f))
    test.endswith(f[0], '/f.lp')

    ground(control)
    test.eq({'a', 'b', 'rule(a,b)'}, {str(a.symbol) for a in control.symbolic_atoms})
    



@test
def reify_until_done(tmp_path):
    control, _ = test_reify_plugin(tmp_path,
    f'#include "{THEORY_FILE}".'"""
    a.
    &rule(b) { a }.
    &rule(c) {} :- b.  % not instantiated until b is True
    """,
    {'#external c.\n', 'b :- a.\n'})
    test.eq({'a', 'b', 'c'}, {str(a.symbol) for a in control.symbolic_atoms})


@test
def reify_with_disappering_atoms(stderr, tmp_path):
    control, _ = test_reify_plugin(tmp_path, """
            a.
            none(notgeel) :-  not geel.  % these cannot
            none(geel)    :-  geel.      % both be true
            rule(geel, a).
            """, {
            'geel :- a.\n'
            })
    symbols = {str(sa.symbol) for sa in control.symbolic_atoms}
    test.contains(symbols, 'none(geel)')
    test.comp.contains(symbols, 'none(notgeel)')


@test
def reify_with_disappering_atoms_in_different_programs(stderr, tmp_path):
    def reify(program_name):
        parts = (('base', ()), (program_name, ()))
        asp_code = """
            none(notgeel) :-  not geel.
            none(geel)    :-  geel.
            rule(geel, a).
            #program test_a.
            a.
            #program test_not_a.
            """
        control, ground = test_reify_plugin(
            tmp_path,
            asp_code,
            {"geel :- a.\n"}, 
            parts=parts)
        ground(control, parts=parts, context=None)
        return {str(sa.symbol) for sa in control.symbolic_atoms}

    symbols = reify('test_a')
    test.contains(symbols, 'none(geel)')
    test.comp.contains(symbols, 'none(notgeel)')

    symbols = reify('test_not_a')
    test.contains(symbols, 'none(notgeel)')
    test.comp.contains(symbols, 'none(geel)')


@test
def reify_with_context_in_plugin(tmp_path):
    class Context:
        @staticmethod
        def zeep():
            return clingo.String("sop")
    control, _ = test_reify_plugin(tmp_path, """
                         b(@zeep).
                         a.
                         rule(geel, a).
                         """,
                         {"geel :- a.\n"},
                         context=Context())
    symbols = {str(sa.symbol) for sa in control.symbolic_atoms}
    test.contains(symbols, 'b("sop")')


@test
def use_proper_logger(tmp_path):
    trace = {}
    def my_logger(code, message):
        trace[code] = message
    try:
        test_reify_plugin(tmp_path, "a", {}, logger=my_logger)
    except RuntimeError as e:
        test.eq('parsing failed', str(e))
    finally:
        code, message = next(iter(trace.items()))
        test.eq(clingo.MessageCode.RuntimeError, code)
        test.endswith(message, " error: syntax error, unexpected EOF\n")


@test
def pass_arguments_properly(tmp_path):
    control, _ = test_reify_plugin(tmp_path, "a(1). a(1/0).", set(), arguments=['--warn', 'no-operation-undefined'])
    test.eq(['a(1)'], list_symbols(control))


@test
def pass_context_properly(tmp_path):
    class Context:
        def f(self):
            return clingo.Number(42)
    control, _ = test_reify_plugin(tmp_path, "a(@f()).", set(), context=Context())
    test.eq(['a(42)'], list_symbols(control))




from .clingo_main_plugin import clingo_main_plugin
from .source_plugin import source_plugin
from .clingo_control_plugin import clingo_control_plugin
from .clingo_syntaxerror_plugin import clingo_syntaxerror_plugin
from .clingo_sequencer_plugin import clingo_sequencer_plugin
from .clingo_defaults_plugin import clingo_defaults_plugin
from .testrunner_plugin import testrunner_plugin
from .clingo_reify_plugin import clingo_reify_plugin, THEORY_PATH
from .insert_plugin_plugin import insert_plugin_plugin
from .stdin_to_tempfile_plugin import stdin_to_tempfile_plugin
from .compound_context_plugin import compound_context_plugin

